/* Copyright (C) 1991, 1994 Aladdin Enterprises.  All rights reserved.

   This file is part of Aladdin Ghostscript.

   Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
   or distributor accepts any responsibility for the consequences of using it,
   or for whether it serves any particular purpose or works at all, unless he
   or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
   License (the "License") for full details.

   Every copy of Aladdin Ghostscript must include a copy of the License,
   normally in a plain ASCII text file named PUBLIC.  The License grants you
   the right to copy, modify and redistribute Aladdin Ghostscript, but only
   under certain conditions described in the License.  Among other things, the
   License requires that the copyright notice and this notice be preserved on
   all copies.
 */

/* gdevrpdl.c */
/* RICOH RPDL driver */
#include "gdevprn.h"
#include "gp.h"		/* for gp_close_printer() */

#define	RPDL_MAX_BITMAP	32767

/* The device descriptors */
private dev_proc_put_params(rpdl_put_params);

private gx_device_procs
rpdl_procs = prn_params_procs(gdev_prn_open, gdev_prn_output_page,
	gdev_prn_close,gdev_prn_get_params,rpdl_put_params);

private dev_proc_print_page(nx100_print_page);

gx_device_printer far_data
gs_nx100f_device = prn_device(rpdl_procs, "nx100f",	/* fixed paper size */
	DEFAULT_WIDTH_10THS_A4, DEFAULT_HEIGHT_10THS_A4, 600, 600,
	0.0, 0.0, 0.0, 0.0,		/* margins */
	1, nx100_print_page);

gx_device_printer far_data
gs_nx100v_device = prn_device(prn_std_procs, "nx100v",	/* variable paper size */
	DEFAULT_WIDTH_10THS_A4, DEFAULT_HEIGHT_10THS_A4, 600, 600,
	0.0, 0.0, 0.0, 0.0,		/* margins */
	1, nx100_print_page);

/* ------ Internal routines ------ */

#define	SUB 0x1a
#define ESC 0x1b
#define	DC2 0x12
#define	EM  0x19

private const char
NX100Init[] = {
	ESC, SUB, 'I',					/* printer reset */
	ESC, DC2, '!', '@', 'R', '0', '0', ESC, ' ',	/* RPDL mode */
	ESC, DC2, 'Y', 'W', ',', '3', ' ',		/* graphics (600dpi) */
	ESC, DC2, 'Y', 'A', '0', '4', ',', '3', ' ',	/* engine (600dpi) */
	ESC, DC2, 'Y', 'B', ',', '2', ' ',		/* max printable area */
	ESC, DC2, '#', '4', ' ',			/* coordination unit(600dpi) */
	ESC, DC2, 'Q', '5', ' ',			/* spaceing unit(600dpi) */
	ESC, DC2, 'M', '1', ' ',			/* normal CR,LF,FF */
	ESC, '!',					/* auto CR-LF disable */
	ESC, EM, 'M',					/* selecting multi feeder tray */
	ESC, DC2, '3', '3', ' ',			/* force exhausting time */
};

typedef struct rpdl_page_setup {
	uint		width, height;	/* [dots] */
	const char	*init;
	bool		landscape;	/* boolean */
} rpdl_page_setup;

private const rpdl_page_setup
RPDLPageSetup[] = {
	{(10500/254 + 2) * 60, (14850/254 + 2) * 60, "\033\022" "52@A6R\033 \033\022" "D1 ", 0},
	{(12850/254 + 2) * 60, (18200/254 + 2) * 60, "\033\022" "52@B6R\033 \033\022" "D1 ", 0},
	{(14850/254 + 2) * 60, (21000/254 + 2) * 60, "\033\022" "52@A5\033 \033\022" "D2 ", 1},
	{(55 + 2) * 60, (85 + 2) * 60, "\033\022" "52@HLR\033 \033\022" "D2 ", 1},
	{(18200/254 + 2) * 60, (25700/254 + 2) * 60, "\033\022" "52@B5\033 \033\022" "D2 ", 1},
	{(85 + 2) * 60, (110 + 2) * 60, "\033\022" "52@LT\033 \033\022" "D2 ", 1},
	{(21000/254 + 2) * 60, (29700/254 + 2) * 60, "\033\022" "52@A4\033 \033\022" "D2 ", 1},
	{(85 + 2) * 60, (140 + 2) * 60, "\033\022" "52@LGR\033 \033\022" "D1 ", 0},
	{(25700/254 + 2) * 60, (36400/254 + 2) * 60, "\033\022" "52@B4R\033 \033\022" "D1 ", 0},
	{(29700/254 + 2) * 60, (42000/254 + 2) * 60, "\033\022" "52@A3R\033 \033\022" "D1 ", 0},
	{(110 + 2) * 60, (170 + 2) * 60, "\033\022" "52@DLR\033 \033\022" "D1 ", 0},
	{max_uint, max_uint, "\033\022" "D1 ", 0}
};

/* run length compression for NX-100 */
private int
rpdl_compress(byte * rp, byte * cp, byte * end_data)
{
	int c = 0, count = 0;
	byte *saved_cp = cp;

	while (rp < end_data) {
		if ((*rp == c) && (count < 256)) {
			if (count == 0) {
				*cp++ = *rp;
			}
			count++;
		} else {
			if (count != 0) {
				*cp++ = count - 1;
				count = 0;
			}
			*cp++ = c = *rp;
		}
		rp++;
	}
	if (count)
		*cp++ = count - 1;

	return (cp - saved_cp);
}

/* Send the page to the printer.  */
private int
nx100_print_page(gx_device_printer * pdev, FILE * pstream)
{
	rpdl_page_setup	*setup_cmd = RPDLPageSetup;
	byte	*row_data, *comp_data;
	int	lnum, ncomp, nrow, line_size = gdev_prn_raster(pdev), nline = pdev->height;

	if (!(row_data = (byte *)gs_malloc(1, line_size*8, "LineBuffer(row)")))
		return_error(gs_error_VMerror);
	if (!(comp_data = (byte *)gs_malloc(1, line_size*12, "LineBuffer(compressed)")))
		return_error(gs_error_VMerror);

	fwrite(NX100Init, (sizeof NX100Init), 1, pstream);
	while ((pdev->width > setup_cmd->width) || (pdev->height > setup_cmd->height)) {
		setup_cmd++;
	}
	fputs(setup_cmd->init, pstream);
	if (setup_cmd->landscape) {
		nline &= ~7;
		for (lnum = 0; lnum < nline; lnum += 8) {
			register int i;
			byte	*in_ptr, *out_ptr, *bgn_data, *end_data;

			gdev_prn_copy_scan_lines(pdev, lnum, comp_data, line_size * 8);
			in_ptr = comp_data;
			out_ptr = row_data + line_size * 8 - 1;
			for (i = 0; i < line_size; i++) {
				gdev_prn_transpose_8x8(in_ptr, line_size, out_ptr, -1);
				out_ptr -= 8;
				in_ptr++;
			}
			bgn_data = row_data;
			end_data = row_data + line_size * 8;

			/* Remove leading 0s. */
			while (end_data > bgn_data && bgn_data[0] == 0)
				bgn_data++;

			/* Remove trailing 0s. */
			while (end_data > bgn_data && end_data[-1] == 0)
				end_data--;

			if ((nrow = end_data - bgn_data) == 0)
				continue;

			/* send the compressed data */
			ncomp = rpdl_compress(bgn_data, comp_data, end_data);
			if(ncomp >= nrow){
				fprintf(pstream, "\033\022" "G3,8,%d,,,%d,%d@",
				nrow, lnum, (bgn_data - row_data));
				fwrite(bgn_data, nrow, 1, pstream);
			}else{
				fprintf(pstream, "\033\022" "G3,8,%d,,4,%d,%d,%d@",
				nrow, lnum, (bgn_data - row_data), ncomp);
				fwrite(comp_data, ncomp, 1, pstream);
			}
		}
	} else {
		for (lnum = 0; lnum < nline; lnum++) {
			byte	*bgn_data, *end_data;

			gdev_prn_copy_scan_lines(pdev, lnum, row_data, line_size);
			bgn_data = row_data;
			end_data = row_data + line_size;

			/* Remove leading 0s. */
			while (end_data > bgn_data && bgn_data[0] == 0)
				bgn_data++;

			/* Remove trailing 0s. */
			while (end_data > bgn_data && end_data[-1] == 0)
				end_data--;

			if ((nrow = end_data - bgn_data) == 0)
				continue;

			/* send the compressed data */
			ncomp = rpdl_compress(bgn_data, comp_data, end_data);
			if(ncomp >= nrow){
				fprintf(pstream, "\033\022" "G3,%d,1,,,%d,%d@",
				nrow*8, (bgn_data - row_data) * 8, lnum);
				fwrite(bgn_data, nrow, 1, pstream);
			}else{
				fprintf(pstream, "\033\022" "G3,%d,1,,4,%d,%d,%d@",
				nrow*8, (bgn_data - row_data) * 8, lnum, ncomp);
				fwrite(comp_data, ncomp, 1, pstream);
			}
		}
	}
	fputc('\f', pstream);

	gs_free(row_data, 1, line_size*8, "LineBuffer(row)");
	gs_free(comp_data, 1, line_size*12, "LineBuffer(compressed)");

	return 0;
}

/* ================ Putting parameters ================ */
#define ppdev ((gx_device_printer *)pdev)

/* Get the color map for a device.  Return true if there is one. */
private bool
param_HWColorMap(gx_device * dev, byte * palette /* 3 << 8 */ )
{
	int depth = dev->color_info.depth;
	int colors = dev->color_info.num_components;

	if (depth <= 8 && colors <= 3) {
		byte *p = palette;
		gx_color_value rgb[3];
		gx_color_index i;
		fill_dev_proc(dev, map_color_rgb, gx_default_map_color_rgb);
		for (i = 0; (i >> depth) == 0; i++) {
			int j;
			if ((*dev_proc(dev, map_color_rgb)) (dev, i, rgb) < 0)
				return false;
			for (j = 0; j < colors; j++)
				*p++ = gx_color_value_to_byte(rgb[j]);
		}
		return true;
	}
	return false;
}

private int
param_check_long(gs_param_list * plist, gs_param_name pname, long value, bool defined)
{
	int code;
	long new_value;
	switch (code = param_read_long(plist, pname, &new_value)) {
	    case 0:
		if (!defined || new_value != value) {
			code = gs_error_rangecheck;
			param_signal_error(plist, pname, code);
		}
		break;
	    default:
		if (param_read_null(plist, pname) == 0)
			return 1;
		param_signal_error(plist, pname, code);
	    case 1:
		break;
	}
	return code;
}
#define param_check_int(plist, pname, ival, defined)\
	param_check_long(plist, pname, (long)(ival), defined)

private int
param_check_bytes(gs_param_list * plist, gs_param_name pname, const byte * str,
			      uint size, bool defined)
{
	int code;
	gs_param_string new_value;
	switch (code = param_read_string(plist, pname, &new_value)) {
	    case 0:
		if (!defined || new_value.size != size ||
		    memcmp((const char *) str, (const char *) new_value.data, size)) {
			code = gs_error_rangecheck;
			param_signal_error(plist, pname, code);
		}
		break;
	    default:
		if (param_read_null(plist, pname) == 0)
			return 1;
		param_signal_error(plist, pname, code);
	    case 1:
		break;
	}
	return code;
}
#define param_check_string(plist, pname, str, defined)\
	param_check_bytes(plist, pname, (const byte *)str, strlen(str), defined)

/* Standard ProcessColorModel values. */
private const char
*pcmsa[] = { "", "DeviceGray", "", "DeviceRGB", "DeviceCMYK" };

/* Put parameters. */
private int
rpdl_put_params(gx_device * pdev, gs_param_list * plist)
{
	int ecode = 0;
	int code;
	int igni;		/* Ignore parameters that only have meaning for printers. */
	bool is_open = pdev->is_open;
	int nci = ppdev->NumCopies;
	bool ncset = ppdev->NumCopies_set;
	bool oof = ppdev->OpenOutputFile;
	bool duplex;
	int duplex_set = -1;
	int width = pdev->width;
	int height = pdev->height;
	gdev_prn_space_params sp = ppdev->space_params;
	gdev_prn_space_params saved_sp = sp;
	gs_param_string ofs;
	gs_param_dict mdict;
	gs_param_float_array hwra;
	gs_param_int_array hwsa;
	gs_param_float_array msa, saved_msa;
	gs_param_float_array ma;
	gs_param_float_array hwma;
	gs_param_float_array mhwra;
	bool ignc = pdev->IgnoreNumCopies;
	gs_param_float_array ibba;
	bool ibbnull = false;
	int colors = pdev->color_info.num_components;
	int depth = pdev->color_info.depth;
	int GrayValues = pdev->color_info.max_gray + 1;
	int RGBValues = pdev->color_info.max_color + 1;
	long ColorValues = 1L << depth;
	gs_param_string cms;
	FILE	*saved_file;

	switch (code = param_read_int(plist, "NumCopies", &nci)) {
	    case 0:
		if (nci < 0) {
			ecode = gs_error_rangecheck;
			param_signal_error(plist, "NumCopies", ecode);
		} else {
			ncset = true;
		}
		break;
	    default:
		if ((code = param_read_null(plist, "NumCopies")) == 0) {
			ncset = false;
			break;
		}
		ecode = code;	/* can't be 1 */
		param_signal_error(plist, "NumCopies", ecode);
	    case 1:
	}

	code = param_read_bool(plist, "OpenOutputFile", &oof);
	if ((code != 0) && (code != 1)) {
		ecode = code;
		param_signal_error(plist, "OpenOutputFile", ecode);
	}

	if (ppdev->Duplex_set >= 0) {	/* i.e., Duplex is supported */
		code = param_read_bool(plist, "Duplex", &duplex);
		if (code == 0) {
			duplex_set = 1;
		} else if (code != 1) {
			if ((code = param_read_null(plist, "Duplex")) == 0) {
				duplex_set = 0;
			} else {
				ecode = code;
				param_signal_error(plist, "Duplex", ecode);
			}
		}
	}

	switch (code = param_read_long(plist, "BufferSpace", &sp.BufferSpace)) {
	    case 0:
		if (sp.BufferSpace >= 10000) break;
		code = gs_error_rangecheck;
	    default:
		ecode = code;
		param_signal_error(plist, "BufferSpace", ecode);
	    case 1:
	}

	switch (code = param_read_long(plist, "MaxBitmap", &sp.MaxBitmap)) {
	    case 0:
		if (sp.MaxBitmap >= 10000) break;
		code = gs_error_rangecheck;
	    default:
		ecode = code;
		param_signal_error(plist, "MaxBitmap", ecode);
	    case 1:
	}

	switch ( code = param_read_int(plist, "BandWidth", &sp.band.BandWidth) ) {
	    case 0:
		if (sp.band.BandWidth >= 0) break;
		code = gs_error_rangecheck;
	    default:
		ecode = code;
		param_signal_error(plist, "BandWidth", ecode);
	    case 1:
	}

	switch ( code = param_read_int(plist, "BandHeight", &sp.band.BandHeight) ) {
	    case 0:
		if (sp.band.BandHeight >= 0) break;
		code = gs_error_rangecheck;
	    default:
		ecode = code;
		param_signal_error(plist, "BandHeight", ecode);
	    case 1:
	}

	switch ( code = param_read_long(plist, "BandBufferSpace", &sp.band.BandBufferSpace) ) {
	    case 0:
		if (sp.band.BandBufferSpace >= 0) break;
		code = gs_error_rangecheck;
	    default:
		ecode = code;
		param_signal_error(plist, "BandBufferSpace", ecode);
	    case 1:
	}

	switch (code = param_read_string(plist, "OutputFile", &ofs)) {
	    case 0:
		if (ofs.size >= prn_fname_sizeof)
			ecode = gs_error_limitcheck;
		else {		/* Check the validity of any % formats. */
			uint i;
			bool pagenum = false;
			for (i = 0; i < ofs.size; ++i)
				if (ofs.data[i] == '%') {
					if (i + 1 < ofs.size && ofs.data[i + 1] == '%')
						continue;
					if (pagenum)	/* more than one %, */
						i = ofs.size - 1;	/* force error */
					pagenum = true;
				      sw:if (++i == ofs.size) {
						ecode = gs_error_rangecheck;
						goto ofe;
					}
					switch (ofs.data[i]) {
					    case ' ':
					    case '#':
					    case '+':
					    case '-':
					    case '0':
					    case '1':
					    case '2':
					    case '3':
					    case '4':
					    case '5':
					    case '6':
					    case '7':
					    case '8':
					    case '9':
					    case 'l':
						goto sw;
					    case 'd':
					    case 'i':
					    case 'u':
					    case 'o':
					    case 'x':
					    case 'X':
						continue;
					    default:
						ecode = gs_error_rangecheck;
						goto ofe;
					}
				}
			break;
		}
		goto ofe;
	    default:
		ecode = code;
	    ofe:param_signal_error(plist, "OutputFile", ecode);
	    case 1:
		ofs.data = 0;
		break;
	}

	/* Read InputAttributes and OutputAttributes just for the type */
	/* check and to indicate that they aren't undefined. */
	code = param_begin_read_dict(plist, "InputAttributes", &mdict, true);
	if (code == 0) {
		param_end_read_dict(plist, "InputAttributes", &mdict);
	} else if (code != 1) {
		ecode = code;
		param_signal_error(plist, "InputAttributes", ecode);
	}
	code = param_begin_read_dict(plist, "OutputAttributes", &mdict, true);
	if (code == 0) {
		param_end_read_dict(plist, "OutputAttributes", &mdict);
	} else if (code != 1) {
		ecode = code;
		param_signal_error(plist, "OutputAttributes", ecode);
	}
	/*
	 * The HWResolution, HWSize, and MediaSize parameters interact in
	 * the following way:
	 *      1. Setting HWResolution recomputes HWSize from MediaSize.
	 *      2. Setting HWSize recomputes MediaSize from HWResolution.
	 *      3. Setting MediaSize recomputes HWResolution from HWSize.
	 * If more than one parameter is being set, we apply these rules
	 * in the order 1, 2, 3.  This does the right thing in the most
	 * common case of setting more than one parameter, namely,
	 * setting both HWResolution and HWSize.
	 */

	switch (code = param_read_float_array(plist, "HWResolution", &hwra)) {
	    case 0:
		if ((hwra.size == 2) && (hwra.data[0] > 0) && (hwra.data[1] > 0)) break;
		code = gs_error_rangecheck;
	    default:
		ecode = code;
		param_signal_error(plist, "HWResolution", ecode);
	    case 1:
		hwra.data = 0;	/* mark as not filled */
	}

	switch (code = param_read_int_array(plist, "HWSize", &hwsa)) {
	    case 0:
		if ((hwsa.size == 2) && (0 < hwsa.data[0]) && (hwsa.data[0] <= RPDL_MAX_BITMAP)
			&& (0 < hwsa.data[1]) && (hwsa.data[1] <= RPDL_MAX_BITMAP)) break;
		code = gs_error_rangecheck;
	    default:
		ecode = code;
		param_signal_error(plist, "HWSize", ecode);
	    case 1:
		hwsa.data = 0;	/* mark as not filled */
	}

	/* .MediaSize takes precedence over PageSize, so */
	/* we read PageSize first. */
	switch (code = param_read_float_array(plist, "PageSize", &msa)) {
	    case 0:
		if ((msa.size == 2) && (msa.data[0] > 0) && (msa.data[1] > 0)) break;
		code = gs_error_rangecheck;
	    default:
		ecode = code;
		param_signal_error(plist, "PageSize", ecode);
	    case 1:
		msa.data = 0;	/* mark as not filled */
	}
	/* Prevent data from being set to 0 if PageSize is specified */
	/* but .MediaSize is not. */
	saved_msa = msa;
	switch (code = param_read_float_array(plist, ".MediaSize", &msa)) {
	    case 0:
		if ((msa.size == 2) && (msa.data[0] > 0) && (msa.data[1] > 0)) break;
		code = gs_error_rangecheck;
	    default:
		ecode = code;
		param_signal_error(plist, ".MediaSize", ecode);
	    case 1:
		msa = saved_msa;	/* mark as not filled */
	}

	switch (code = param_read_float_array(plist, "Margins", &ma)) {
	    case 0:
		if (ma.size == 2) break;
		code = gs_error_rangecheck;
	    default:
		ecode = code;
		param_signal_error(plist, "Margins", ecode);
	    case 1:
		ma.data = 0;	/* mark as not filled */
	}

	switch (code = param_read_float_array(plist, ".HWMargins", &hwma)) {
	    case 0:
		if (hwma.size == 4) break;
		code = gs_error_rangecheck;
	    default:
		ecode = code;
		param_signal_error(plist, ".HWMargins", ecode);
	    case 1:
		hwma.data = 0;	/* mark as not filled */
	}

	/* MarginsHWResolution cannot be changed, only checked. */
	switch (code = param_read_float_array(plist, ".MarginsHWResolution", &mhwra)) {
	    case 0:
		if ((mhwra.size == 2)
		    && (mhwra.data[0] == pdev->MarginsHWResolution[0])
		    && (mhwra.data[1] == pdev->MarginsHWResolution[1])) break;
		code = gs_error_rangecheck;
	    default:
		ecode = code;
		param_signal_error(plist, ".MarginsHWResolution", ecode);
	    case 1:
		mhwra.data = 0;	/* mark as not filled */
	}

	code = param_read_bool(plist, ".IgnoreNumCopies", &ignc);
	if ((code != 0) && (code != 1)) {
		ecode = code;
		param_signal_error(plist, ".IgnoreNumCopies", ecode);
	}

	/* Ignore parameters that only have meaning for printers. */
	code = param_read_int(plist, "%MediaSource", &igni);
	if ((code != 0) && (code != 1)) {
		ecode = code;
		param_signal_error(plist, "%MediaSource", ecode);
	}

	code = param_read_int(plist, "%MediaDestination", &igni);
	if ((code != 0) && (code != 1)) {
		ecode = code;
		param_signal_error(plist, "%MediaDestination", ecode);
	}

	switch (code = param_read_float_array(plist, "ImagingBBox", &ibba)) {
	    case 0:
		if (ibba.size != 4 || ibba.data[2] < ibba.data[0] || ibba.data[3] < ibba.data[1]) {
			ecode = gs_error_rangecheck;
			param_signal_error(plist, "ImagingBBox", ecode);
			ibba.data = 0;
		}
		break;
	    default:
		if ((code = param_read_null(plist, "ImagingBBox")) == 0) {
			ibbnull = true;
			ibba.data = 0;
			break;
		}
		ecode = code;	/* can't be 1 */
		param_signal_error(plist, "ImagingBBox", ecode);
	    case 1:
		ibba.data = 0;
	}

	/* Now check nominally read-only parameters. */
	if ((code = param_check_string(plist, "OutputDevice", pdev->dname, true)) < 0)
		ecode = code;
	if ((code = param_check_string(plist, "ProcessColorModel", pcmsa[colors], colors != 0)) < 0)
		ecode = code;
	if ((code = param_check_string(plist, "Name", pdev->dname, true)) < 0)
		ecode = code;
	if ((code = param_check_int(plist, "Colors", colors, true)) < 0)
		ecode = code;
	if ((code = param_check_int(plist, "BitsPerPixel", depth, true)) < 0)
		ecode = code;
	if ((code = param_check_int(plist, "GrayValues", GrayValues, true)) < 0)
		ecode = code;
	if ((code = param_check_long(plist, "PageCount", pdev->PageCount, true)) < 0)
		ecode = code;
	if ((code = param_check_int(plist, "RedValues", RGBValues, colors > 1)) < 0)
		ecode = code;
	if ((code = param_check_int(plist, "GreenValues", RGBValues, colors > 1)) < 0)
		ecode = code;
	if ((code = param_check_int(plist, "BlueValues", RGBValues, colors > 1)) < 0)
		ecode = code;
	if ((code = param_check_long(plist, "ColorValues", ColorValues, colors > 1)) < 0)
		ecode = code;
	if (param_read_string(plist, "HWColorMap", &cms) != 1) {
		byte palette[3 << 8];
		if (param_HWColorMap(pdev, palette))
			code = param_check_bytes(plist, "HWColorMap", palette,
						 colors << depth, true);
		else
			code = param_check_bytes(plist, "HWColorMap", 0, 0, false);
		if (code < 0)
			ecode = code;
	}
	if ((code = param_check_int(plist, "TextAlphaBits",
	   (*dev_proc(pdev, get_alpha_bits)) (pdev, go_text), true)) < 0)
		ecode = code;

	if ((code = param_check_int(plist, "GraphicsAlphaBits",
	(*dev_proc(pdev, get_alpha_bits)) (pdev, go_graphics), true)) < 0)
		ecode = code;

	/* We must 'commit', in order to detect unknown parameters, */
	/* even if there were errors. */
	code = param_commit(plist);
	if (ecode < 0)
		return ecode;
	if (code < 0)
		return code;

	/* Now actually make the changes. */

	if (hwra.data != 0) {
		pdev->HWResolution[0] = hwra.data[0];
		pdev->HWResolution[1] = hwra.data[1];
		pdev->width = pdev->MediaSize[0] * hwra.data[0] / 72.0 + 0.5;
		pdev->height = pdev->MediaSize[1] * hwra.data[1] / 72.0 + 0.5;
	}
	if (hwsa.data != 0) {
		pdev->width = hwsa.data[0];
		pdev->height = hwsa.data[1];
		pdev->MediaSize[0] = hwsa.data[0] * 72.0 / pdev->HWResolution[0];
		pdev->MediaSize[1] = hwsa.data[1] * 72.0 / pdev->HWResolution[1];
	}
	if (msa.data != 0) {
		pdev->MediaSize[0] = msa.data[0];
		pdev->MediaSize[1] = msa.data[1];
		pdev->HWResolution[0] = pdev->width * 72.0 / msa.data[0];
		pdev->HWResolution[1] = pdev->height * 72.0 / msa.data[1];
	}
	if (ma.data != 0) {
		pdev->Margins[0] = ma.data[0];
		pdev->Margins[1] = ma.data[1];
	}
	if (hwma.data != 0) {
		pdev->HWMargins[0] = hwma.data[0];
		pdev->HWMargins[1] = hwma.data[1];
		pdev->HWMargins[2] = hwma.data[2];
		pdev->HWMargins[3] = hwma.data[3];
	}
	pdev->IgnoreNumCopies = ignc;
	if (ibba.data != 0) {
		pdev->ImagingBBox[0] = ibba.data[0];
		pdev->ImagingBBox[1] = ibba.data[1];
		pdev->ImagingBBox[2] = ibba.data[2];
		pdev->ImagingBBox[3] = ibba.data[3];
		pdev->ImagingBBox_set = true;
	} else if (ibbnull) {
		pdev->ImagingBBox_set = false;
	}
	ppdev->NumCopies = nci;
	ppdev->NumCopies_set = ncset;
	saved_file = ppdev->file;
	ppdev->OpenOutputFile = false;
	ppdev->file = NULL;
	if (duplex_set >= 0) {
		ppdev->Duplex = duplex;
		ppdev->Duplex_set = duplex_set;
	}
	/* If necessary, free and reallocate the printer memory. */
	if ( memcmp(&sp, &saved_sp, sizeof(sp)) != 0 ||
	    pdev->width != width || pdev->height != height) {
		if (is_open)
			gdev_prn_close(pdev);
		ppdev->space_params = sp;
		if (is_open) {
			ecode = code = gdev_prn_open(pdev);
			if (code < 0) {		/* Try to put things back as they were. */
				ppdev->space_params = saved_sp;
				pdev->width = width;
				pdev->height = height;
				pdev->MediaSize[0] = width * 72.0 / pdev->HWResolution[0];
				pdev->MediaSize[1] = height * 72.0 / pdev->HWResolution[1];
				code = gdev_prn_open(pdev);
				if (code < 0) {		/* Disaster!  We can't get back. */
					pdev->is_open = false;
					return code;
				}
				return ecode;
			}
		}
	}
	ppdev->file = saved_file;
	ppdev->OpenOutputFile = oof;
	if (ofs.data != 0 &&
	    bytes_compare(ofs.data, ofs.size,
		    (const byte *) ppdev->fname, strlen(ppdev->fname))) {
		/* Close the file if it's open. */
		if (ppdev->file != NULL && ppdev->file != stdout)
			gp_close_printer(ppdev->file, ppdev->fname);
		ppdev->file = NULL;
		memcpy(ppdev->fname, ofs.data, ofs.size);
		ppdev->fname[ofs.size] = 0;
	}
	/* If the device is open and OpenOutputFile is true, */
	/* open the OutputFile now.  (If the device isn't open, */
	/* this will happen when it is opened.) */
	if (pdev->is_open && oof) {
		code = gdev_prn_open_printer(pdev, 1);
		if (code < 0)
			return code;
	}
	return 0;
}
